# coding: utf-8

"""
    Pearson Athena

     Module containing a SQL query builder for Athena and an Athena connection that simplifies interactions with AWS Athena.   # noqa: E501
"""


from setuptools import setup, find_packages  # noqa: H301

NAME = "prsn-athena"
VERSION = "0.2.0"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["boto3 >= 1.14.63", "pandas >= 1.0.5", "requests >= 2.23.0"]

setup(
    name=NAME,
    version=VERSION,
    description="Pearson Athena",
    author="Pearson Cloud Management",
    author_email="io-pearson-cloud-management@pearson.com",
    url="",
    keywords=["Pearson", "PCM", "Athena"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
     Module containing a SQL query builder for Athena and an Athena connection that simplifies interactions with AWS Athena.   # noqa: E501
    """
)
